#include "simulation.h"
#include "random.h"
#include "tcoopt.cpp"
#include <time.h>
#include <math.h>

#define POCET_DELNIKU	200
#define POCET_SKUPIN	4
#define POCET_HOSPOD	8
#define POCET_STOLU		8
#define POCET_MIST		4
#define DOBA_OBSLUHY	5
#define MAX_PIV			10

//-------------------------------SHO demo-------------------------------- 

class THospoda; class TStul; class TDelnik;

THospoda *hospoda[POCET_HOSPOD];
TIME posledni;
int piv[MAX_PIV + 1];
int counterStul = 0;
int counterDelnik = 0;


class THospoda : public CProcess {               // process servicing requests
public:
  TStul *stul[POCET_STOLU];
  CHead *VIP;
  CHead *socky;
  int skupina;

  THospoda();
  virtual void Run();                            // service's behaviour
  TStul * getStul(int skupina);
};

class TDelnik : public CProcess {               // request in the queueing system
  const static int pPrahDalsiPivo = 1;
  const static int pPrahDalsiHospoda = 2;
  const static int pMaxPiv = MAX_PIV;

public:
  int num, vHospode, skupina, vypitoPiv;
  TDelnik();
  virtual void Run();                            // request's behaviour
  bool dalsiHospoda();
  bool dalsiPivo(int lidi);
  void napijSe();
};

class TStul : public CProcess {               // request in the queueing system
public:
  int volno, skupina, num;
  CHead *prisedici;
  THospoda *hospoda;

  TStul(THospoda *hospoda);
  virtual void Run();                            // request's behaviour
  int pocetLidi();
  void zaradDoFronty();
};

TStul::TStul(THospoda *h) {
  hospoda = h;
  volno = POCET_MIST;
  prisedici = new CHead;
  skupina = 0;
  num = ++counterStul;
}

void TStul::Run() {
  while(true) {
	while(!prisedici->Empty()) {
	  printf("Stul %d se radi do fronty: %f\n", num, Time());
	  zaradDoFronty();
      Passivate();
	  printf("Stul %d obslouzen: %f\n", num, Time());
	  Hold(uniform(20,30));
	  printf("Stul %d se napil: %f\n", num, Time());
	  TDelnik *d = (TDelnik *) prisedici->First();
	  while(d) {
		d->napijSe();
		d = (TDelnik *) d->Suc();
	  }
	  bool odesel = true;
	  while(odesel) {
	    odesel = false;
	    d = (TDelnik *) prisedici->First();
	    while(d) {
          TDelnik *v = d;
		  d = (TDelnik *) d->Suc();
		  if(!v->dalsiPivo(pocetLidi())) {
			v->Out();
			odesel = true;
			volno++;
			v->ActivateAt(Time ());
		  }
	    }
	  }
	}
	skupina = 0;
	Passivate();
  }
}

int TStul::pocetLidi() {
  return POCET_MIST - volno;
}

void TStul::zaradDoFronty() {
  CHead *fronta;
  if(skupina == hospoda->skupina) {
    fronta = hospoda->VIP;
  } else {
    fronta = hospoda->socky;
  }
  if(hospoda->VIP->Empty() && hospoda->socky->Empty()) {
    Into(fronta);
	hospoda->ActivateAt(Time());
  } else {
    Into(fronta);
  }
}

TDelnik::TDelnik() {
  vypitoPiv = 0;
  num = ++counterDelnik;
  skupina = randint(1, POCET_SKUPIN);
}

void TDelnik::Run() {
  printf("Delnik %03d odchazi z tovarny: %f\n", num, Time());
  while(dalsiHospoda()) {
    vHospode = randint(0, POCET_HOSPOD - 1);
	printf("Delnik %03d jde do hospody %d: %f\n", num, vHospode, Time());
	Hold(uniform(10,30));
	TStul *stul = (hospoda[vHospode])->getStul(skupina);
	if(stul) {
      printf("Delnik %03d si priseda v hospode %d ke stolu %d, volno %d\n", num, vHospode, stul->num, stul->volno);
      if(stul->prisedici->Empty()) {
	    Into(stul->prisedici);
        stul->skupina = skupina;
	    stul->ActivateAt(Time ());
      } else {
	    Into(stul->prisedici);
      }
      stul->volno--;
	  Passivate();
	}
	else printf("V hospode %d neni pro delnika %03d volne misto: %f\n", vHospode, num, Time());
  }
  printf("Delnik %03d jde domu, ma %d vypitych piv: %f\n", num, vypitoPiv, Time());
  piv[vypitoPiv]++;
  posledni = Time();
}

bool TDelnik::dalsiHospoda() {
  return (pMaxPiv-vypitoPiv)*normal(0.5,0.75)>pPrahDalsiHospoda;
}

bool TDelnik::dalsiPivo(int lidi) {
  return (pMaxPiv-vypitoPiv )*lidi*normal(0.5,0.75)>pPrahDalsiPivo;
}

void TDelnik::napijSe() {
  printf("Delnik %03d se napil: %f\n", num, Time());
  vypitoPiv++;
}

THospoda::THospoda() {
  for(int i = 0; i < POCET_STOLU; i++) {
    stul[i]  = new TStul(this);			              // creating instance of queue
  }
  VIP =  new CHead();
  socky =  new CHead();
  skupina = randint(1, POCET_SKUPIN);
}

void THospoda::Run() {
  CProcess* P;                                       // temp. variable
  while(true) {
     if (VIP->Empty() && socky->Empty()) {
       Passivate();                                 // service is waiting for incoming requests
     } else if(!VIP->Empty()) {
       P=(CProcess*)VIP->First();                
       P->Out();                                      // removing the request from the queue
       Hold(DOBA_OBSLUHY);                             // servicing time (mean value 50 )
       P->ActivateAt (Time());                        // executing request
     } else {
       P=(CProcess*) socky->First();                
       P->Out();                                      // removing the request from the queue
       Hold(DOBA_OBSLUHY);                             // servicing time (mean value 50 )
       P->ActivateAt (Time());                        // executing request
     }
  }
}

TStul * THospoda::getStul(int s) {
  for(int i = 0; i < POCET_STOLU; i++) {
	if((stul[i]->skupina == s) && (stul[i]->pocetLidi() != POCET_MIST)) return stul[i];
  }
  for(int i = 0; i < POCET_STOLU; i++) {
	if(stul[i]->pocetLidi() == 0) return stul[i];
  }
  return NULL;
}

void CSimulation::Run()								// main process
{
  printf("--- Demonstration of SIMULATION module -- time %f\n", Time());

  srand( (unsigned)time( NULL ) );					// inicialization on random number generator
  for(int i = 0; i < POCET_HOSPOD; i++) {
	hospoda[i] = new THospoda();					// creating service process
  }
  for(int i = 0; i < POCET_DELNIKU; i++) {
	(new TDelnik)->ActivateDelay((float)i/2);       // creating and scheduling 1st request
  }
  for(int i = 0; i < MAX_PIV; i++) {
    piv[i] = 0;
  }
  
  Hold(60*24*7);	//tyden
  printf("End of simulation --- time: %f\n\n\n", Time ());
  printf("Histogram:\n");
  printf(" 0  |  1  |  2  |  3  |  4  |  5  |  6  |  7  |  8  |  9  |  10\n");
  printf("---------------------------------------------------------------\n");
  printf("%3d | %3d | %3d | %3d | %3d | %3d | %3d | %3d | %3d | %3d | %3d\n\n", piv[0], piv[1], piv[2], piv[3], piv[4], piv[5], piv[6], piv[7], piv[8], piv[9], piv[10]);
  float sum = 0;
  for(int i = 0; i <= MAX_PIV; i++) {
    sum += (piv[i] * i);
  }
  printf("Prumerne piv na delnika: %.3f\n", sum / POCET_DELNIKU);
  TIME l = (posledni/60);
  int min = (int) ((l - floor(l)) * 60);
  printf("Cas: %.0f:%02d\n", 17 + l, min);
  exit(0);
};